/***************************************************************************** 
 
	File		: reader.c 
	Last modified	: 06.07.95 
			  set the end correct 
 
	Author		: Roman WIlli, romwilli@iiic.ethz.ch 
 
****************************************************************************** 
	Description	: This module scans a HTML-File and starts 
			  HT_Transform 
 
****************************************************************************** 
	includes 
*****************************************************************************/ 
 
#include <stdio.h> 
#include <stdlib.h> 
#include <string.h> 
#include <sys/types.h> 
#include <sys/stat.h> 
 
#include "reader.h" 
#include "HTBlinds.h"
 
/***************************************************************************** 
	constants 
*****************************************************************************/ 
 
#define MANY    '.'     /* MANY = '.'   any string		       	*/ 
#define NONE    '_'     /* NONE = '_'   no matching		    	*/ 
#define NOT     '~'     /* NOT  = '~'   not the next char		*/ 
#define ANY     '*'     /* ANY  = '*'   any char			*/ 
#define DIGIT   '1'     /* DIGIT= '1'   number 0-9			*/ 
 
#define MAX_BUF 1023 	/* size of the array buf	*/
#define MAXLEN_STR 60 	/* size of the array string	*/
 
#define IMGT	1	/* private symboldefinitions 	*/
#define END 	2	/* 	"	"	 	*/
#define BR  	3 	/*	"	"		*/
#define ENDH 	4	/*	"	"		*/
#define STRI0	5	/*	"	"		*/
#define STRI1	6	/*	"	"		*/
#define P1	7	/*	"	"		*/

#define MAX_SYM 54	/* size of the array symbol	*/

/* definitions of the HTML-Symbols */

#define ADR0SYM  "<address>" 
#define ADR1SYM  "</address>" 
#define BLK0SYM  "<blockquote>" 
#define BLK1SYM  "</blockquote>" 
#define TT0SYM   "<tt>"  
#define TT1SYM   "</tt>" 
#define B0SYM    "<b>" 
#define B1SYM    "</b>" 
#define I0SYM    "<i>" 
#define I1SYM    "</i>" 
#define U0SYM    "<u>" 
#define U1SYM    "</u>" 
#define EM0SYM   "<em>" 
#define EM1SYM   "</em>" 
#define STR0SYM  "<strong>" 
#define STR1SYM  "</strong>" 
#define DFN0SYM  "<dfn>" 
#define DFN1SYM  "</dfn>" 
#define CIT0SYM  "<cite>" 
#define CIT1SYM  "</cite>" 
#define VAR0SYM  "<var>" 
#define VAR1SYM  "</var>" 
#define TIT0SYM  "<h1.>" 
#define TIT1SYM  "</h1.>" 
#define IMGSYM   "<img.>" 
#define RADSYM   "<input.type=.radio.>" 
#define CHKSYM   "<input.type=.checkbox.>" 
#define BUTSYM   "<input.type=.submit.>" 
#define RESSYM   "<input.type=.reset.>"
#define IMGTSYM  "<input.type=.image.>"
#define EDTSYM   "<input.>" 
#define CBBSYM   "<select.>" 
#define TXTSYM   "<textarea.>" 
#define LINKSYM0 "<a.href=.>" 
#define LINKSYM1 "</a>" 
#define BDYSYM   "<body.>" 
#define ENDSYM   "</body>" 
#define PSYM     "<p>"
#define P1SYM	 "<p .>" 
#define BRSYM    "<br>" 
#define FT0SYM	 "<font.>" 
#define FT1SYM	 "</font>" 
#define CEN0SYM	 "<center.>" 
#define CEN1SYM	 "</center>" 
#define ENDHSYM	 "</html>"
#define S0SYM	 "<s>"
#define S1SYM	 "</s>"
#define BIG0SYM	 "<big>"
#define BIG1SYM	 "</big>"
#define SML0SYM	 "<small>"
#define SML1SYM	 "</small>"
#define STRI0SYM "<strike>"
#define STRI1SYM "</strike>"
 
/***************************************************************************** 
	variables & datatypes 
*****************************************************************************/ 
 
typedef char string[MAXLEN_STR]; 	/* typedefinition for string        */
 
static string symbol[MAX_SYM];		/* array with the HTML-symbols	    */
static char buf[MAX_BUF]; 		/* buffer for the HTML-strings	    */
static char help_buf[MAX_BUF];		/* help-buffer for titels and links */
static char help_tit[MAX_BUF];		/* help-buffer for titels if ther
					   is found a link inside the titel */
static int awaited_sym=0;		/* number of the awaited symbol :
					   0    -> no symbol is awaited
					   LNK1 -> end of link is awaited
					   TIT1 -> end of titel is awaited  */
static int begin_set=0;			/* begin setting flag :
					   0 -> begin not set
					   1 -> begin set		    */ 
static int end_set=0;                	/* end setting flag
					   0 -> end not set
					   1 -> end set			    */ 
static int in_buf_pos=0;		/* position in the HTML-buffer b    */
static int out_buf_pos=0;		/* position in the buffer buf       */
static sym=0;				/* number of the current symbol     */
 
/*static FILE* fp_scan;	*/		/* filepointer for the logfile
					   "scan.log"                       */
/*static int file_open=0; */		/* file open flag :
					   0 -> logfile not open
					   1 -> logfile open                */
 
/***************************************************************************** 
	procedures & functions 
*****************************************************************************/ 
 
/***************************************************************************** 
	Transform() 
*****************************************************************************/ 
int Transform(int sym, char *out,HTStream* me) 
{ 
/*	on entry    : sym -> recognized symbol 
		      out -> string to get transformed 
		      me  -> HtBlinds Stream 
	on exit     : returnvalue of transform 
	description : starts transform
	variables   : -
*/ 
  /*output in logfile
  fprintf(fp_scan,"%i\t%s@\n",sym, out);*/
 
  return transform(sym,out,me); 
} /* end Transform */ 
 
/***************************************************************************** 
	InitSym() 
*****************************************************************************/ 
void InitSym() 
/*	on entry    : void 
	on exit     : void 
	description : initials the pattern-strings
	variables   : -
*/ 
{ 
  strcpy(symbol[ADR0],ADR0SYM);         strcpy(symbol[ADR1],ADR1SYM); 
  strcpy(symbol[BLK0],BLK0SYM);         strcpy(symbol[BLK1],BLK1SYM); 
  strcpy(symbol[TT0],TT0SYM);           strcpy(symbol[TT1],TT1SYM); 
  strcpy(symbol[B0],B0SYM);             strcpy(symbol[B1],B1SYM); 
  strcpy(symbol[I0],I0SYM);             strcpy(symbol[I1],I1SYM); 
  strcpy(symbol[U0],U0SYM);             strcpy(symbol[U1],U1SYM); 
  strcpy(symbol[EM0],EM0SYM);           strcpy(symbol[EM1],EM1SYM); 
  strcpy(symbol[STR0],STR0SYM);         strcpy(symbol[STR1],STR1SYM); 
  strcpy(symbol[DFN0],DFN0SYM);         strcpy(symbol[DFN1],DFN1SYM); 
  strcpy(symbol[CIT0],CIT0SYM);         strcpy(symbol[CIT1],CIT1SYM); 
  strcpy(symbol[VAR0],VAR0SYM);         strcpy(symbol[VAR1],VAR1SYM); 
  strcpy(symbol[TIT],TIT0SYM);          strcpy(symbol[TIT1],TIT1SYM); 
  strcpy(symbol[IMG],IMGSYM);           strcpy(symbol[RAD],RADSYM); 
  strcpy(symbol[CHK],CHKSYM);           strcpy(symbol[BUT],BUTSYM);
  strcpy(symbol[IMGT],IMGTSYM);		strcpy(symbol[EDT],EDTSYM);
  strcpy(symbol[CBB],CBBSYM); 
  strcpy(symbol[LNK],LINKSYM0);         strcpy(symbol[LNK1],LINKSYM1); 
  strcpy(symbol[BDY],BDYSYM);           strcpy(symbol[END],ENDSYM); 
  strcpy(symbol[TXT],TXTSYM);           strcpy(symbol[RES],RESSYM); 
  strcpy(symbol[P],PSYM);               strcpy(symbol[BR],BRSYM); 
  strcpy(symbol[FT0],FT0SYM);		strcpy(symbol[FT1],FT1SYM); 
  strcpy(symbol[CEN0],CEN0SYM);		strcpy(symbol[CEN1],CEN1SYM); 
  strcpy(symbol[ENDH],ENDHSYM); 	
  strcpy(symbol[S0],S0SYM);		strcpy(symbol[S1],S1SYM);
  strcpy(symbol[BIG0],BIG0SYM);		strcpy(symbol[BIG1],BIG1SYM);
  strcpy(symbol[SML0],SML0SYM);		strcpy(symbol[SML1],SML1SYM);
  strcpy(symbol[STRI0],STRI0SYM);	strcpy(symbol[STRI1],STRI1SYM);
 
} /* end InitSym */ 
 
/***************************************************************************** 
	ReadBlanc() 
*****************************************************************************/ 
int ReadBlanc(int j, char *comp) 
/*	on entry      : j    -> actual position in the buffer 
			        of the HTML-document 
			comp -> the buffer of the HTML-document 
	on exit       : new actual position in the buffer 
	description   :	reads the blancs until a another char 
			is found
	variables     : -
			 
*/ 
{ 
  while ((j<strlen(comp))&&(comp[j]==' ')) { 
    j++; 
  } 
  return j; 
} /* end ReadBlanc */ 
 
/***************************************************************************** 
	Compare() 
*****************************************************************************/ 
int Compare(char *muster, char *comp) 
/*	on entry      : muster -> pattern-string 
			comp   -> read string 
	on exit       : 0 -> strings are equal 
			1 -> strings are not equal 
	description   :	compares a symbol-string with 
			a read string from the HTML-document
	variables     : i -> counter for the string muster
		        j -> counter for the string comp
*/ 
{ 
  int i=0,j=0; 
 
  while ((i<strlen(muster))&&(j<strlen(comp))) { 
    switch(muster[i]) { 
      case NONE : 
	return 1; 
	break; 
      case NOT: 
	if (muster[i+1]!=' ') {
	  j=ReadBlanc(j,comp);
        }
	if (muster[i+1]== tolower(comp[j])) { 
	  return 1; 
	} else {  
	  i=i+2; j++; 
	} 
	break; 
      case DIGIT: 
	j=ReadBlanc(j,comp); 
	if ((comp[j]>='0')&&(comp[j]<='9')) { 
	  i++; j++; 
	} else { 
	  return 1; 
	} 
	break; 
      case ANY: 
	j=ReadBlanc(j,comp); 
	i++; j++; 
	break; 
      case MANY: 
	i++; 
	if (i== strlen(muster)) { 
	  return 0; 
	} else { 
	  for (;;) { 
	    while ((j<strlen(comp))&&(muster[i]!=tolower(comp[j]))) { 
	      j++; 
	    } /* end while */ 
	    if (j==strlen(comp)) { 
	      return 1; 
	    } else { 
	      if (Compare(muster+i,comp+j)==0) { 
		return 0; 
	      } else { 
		j++; 
	      } 
	    } 
	  } /* end for */ 
	} 
	break; 
      default:
        if (muster[i]!=' ') {
	  j=ReadBlanc(j,comp);
        } 
	if (muster[i]==tolower(comp[j])) { 
	  i++; j++; 
	} else { 
	  return 1; 
	} 
	break; 
    } /* switch */ 
  } /* while */ 
  if ((i==strlen(muster))&&(j==strlen(comp))) { 
    return 0; 
  } else { 
    return 1; 
  } 
} /* end Compare */ 
 
/***************************************************************************** 
	GetSym() 
*****************************************************************************/ 
int GetSym() 
/*	on entry      : void			 
	on exit       : if one symbolstring matches -> symbolnumber 
			else                        -> PUT
	description   :	compares al symbolstrings with the buffer
	variables     : i -> counter for the symbols
*/ 
{ 
  int i=1;
 
  while ((i<MAX_SYM)&&(Compare(symbol[i],buf)!=0)) { 
    i++; 
  } 
  if (i==MAX_SYM) { 
    return PUT; 
  } else { 
    return i; 
  } 
} /* end GetSym */ 

/***************************************************************************** 
	Delete_buf() 
*****************************************************************************/ 
void Delete_buf(char* b,int n)
/*	on entry      : b -> buffer with strings from the HTML-document
			n -> actual buffer pos
	on exit       :	void	 
	description   :	delete the content of the buffer
	variables     : i -> counter for the content of the buffer
*/ 
{
  int i;

  for (i=0 ; i<= n ; i++) {
    b[i]='\0';
  }
}
 
/***************************************************************************** 
	GetString() 
*****************************************************************************/ 
int GetString(CONST char* b) 
/*	on entry      : buffer with strings from 
			the HTML-document 
	on exit       :	0 -> valid string, not end of buffer b   
			     (out_buf_pos == 0 / in_buf_pos == cur. pos.) 
		      	1 -> valid string, end of buffer b       
                             (out_buf_pos == 0 / in_buf_pos == 0) 
                        2 -> not valid string, end of buffer b	
                             (out_buf_pos == cur. pos. / in_buf_pos == 0)
			3 -> buffer buf overflow
			     (out_buf_pos == 0 / in_buf_pos == cur. pos.)	
	 
	description   :	read the buffer until end of buffer or 
			a valid string is read. 
			valid string :  "<......>" 
					".......>" 
					".......<" 
			(b.w. '.' = any char except "<" or ">" )
	variables     : -
*/ 
{ 

  if (out_buf_pos<MAX_BUF-1 && in_buf_pos<strlen(b)) { 
    buf[out_buf_pos]=b[in_buf_pos];
    in_buf_pos++; out_buf_pos++; 
  } 
  while (out_buf_pos<MAX_BUF-1 && in_buf_pos<strlen(b) && 
	 b[in_buf_pos]!='<' && 
	 b[in_buf_pos]!='>') { 
    buf[out_buf_pos]=b[in_buf_pos]; 
    in_buf_pos++; out_buf_pos++; 
  } 
  if (in_buf_pos>=strlen(b)) { 
    buf[out_buf_pos]='\0';
    Delete_buf(b,in_buf_pos); 
    in_buf_pos=0; 
    if (buf[0]!='<') { 
      out_buf_pos=0;  
      return 1; 
    } else {  
      return 2; 
    } 
  } else if (out_buf_pos>=MAX_BUF-1) { 
    buf[out_buf_pos]='\0'; 
    out_buf_pos=0;
    return 3; 
  } else { 
    if (b[in_buf_pos]=='>') { 
      buf[out_buf_pos]='>'; out_buf_pos++; 
      if (out_buf_pos<MAX_BUF) { 
        buf[out_buf_pos]='\0'; 
      } 
      if (in_buf_pos<strlen(b)) { 
        in_buf_pos++; 
      } 
      out_buf_pos=0; 
      return 0; 
    } else {
      buf[out_buf_pos]='\0'; 
      out_buf_pos=0; 
      return 0;
    } 
  }  
} /* end GetString */ 
 
/***************************************************************************** 
	SetBegin() 
*****************************************************************************/ 
void SetBegin(HTStream* me) 
/*	on entry      : me -> HTBlinds-Stream 
	on exit       : void 
	description   :	set the begin symbols
	variables     : -
*/ 
{ 
  if (begin_set==0) { 
    Transform(BEG,"\0",me); 
    begin_set=1; 
  } 
} /* end SetBegin */ 
 
/***************************************************************************** 
	Reader() 
*****************************************************************************/ 
void Reader(int sym,HTStream* me,char* out_string) 
/*	on entry      : sym 	   -> symbolnumber 
			me  	   -> HTBlinds-Stream 
			out_string -> string to get transformed 
	on exit       : void 
	description   :	starts transform
	variables     : -
*/ 
{ 
  switch(sym) { 
    case BDY : 
      Transform(BDY,out_string,me); 
      SetBegin(me); 
      break; 
    case RAD: 
    case CHK: 
    case TT1: 
    case BUT: 
    case EDT: 
    case CBB: 
      SetBegin(me); 
      Transform(sym,"\0",me); 
      Transform(PUT,out_string,me); 
      break;
    case IMGT:
      SetBegin(me); 
      Transform(BUT,"\0",me); 
      Transform(PUT,out_string,me); 
      break;
    case TXT: 
      SetBegin(me); 
      Transform(EDT,"\0",me); 
      Transform(PUT,out_string,me); 
      break; 
    case RES: 
      SetBegin(me); 
      Transform(BUT,"\0",me); 
      Transform(PUT,out_string,me); 
      break; 
    case LNK1: 
    case TIT1:
      SetBegin(me);
      break; 
    case END: 
      Transform(EDC,"\0",me); 
      Transform(LLS,"\0",me); 
      Transform(ELL,"\0",me); 
      Transform(TLS,"\0",me); 
      Transform(PUT,out_string,me); 
      Transform(BLU,"\0",me); 
      end_set=1; 
      break;
    case BR: 
    case PUT: 
    case ENDH: 
      Transform(PUT,out_string,me);
      break;
    case STRI0:
      SetBegin(me);
      Transform(S0,out_string,me);
      break;
    case STRI1:
      SetBegin(me);
      Transform(S1,out_string,me);
      break;    
    default: 
      SetBegin(me); 
      Transform(sym,out_string,me); 
      break; 
   } 
} /* end Reader */ 
 
/***************************************************************************** 
	Reset() 
*****************************************************************************/ 
void Reset() 
/*	on entry      : void			 
	on exit       : void			 
	description   :	reset the HT_Reader
	variables     : -			 
*/ 
{ 
  /* close logfile
  fclose(fp_scan); file_open=0;*/
 
  awaited_sym=0; 
  begin_set=0; 
  end_set=0; 
  out_buf_pos=0; 
  in_buf_pos=0; 
  sym=0; 
  buf[0]= '\0'; 
  help_buf[0]= '\0'; 
  help_tit[0]= '\0'; 
} /* end Reset */ 
 
/***************************************************************************** 
	Blinds_sendEOF() 
*****************************************************************************/ 
void Blinds_sendEOF(HTStream* me) 
/*	on entry      : me -> HTBlinds-Stream			 
	on exit       : void			 
	description   :	set the end symbols if in the HTML-Document
			isn't the endstring "</body>"
	variables     : -		 
*/ 
{ 
  if (end_set==0) { 
    Transform(EDC,"\0",me); 
    Transform(LLS,"\0",me); 
    Transform(ELL,"\0",me); 
    Transform(TLS,"\0",me); 
    Transform(BLU,"\0",me);  
  } 
  Reset(); 
  close_transform(); 
} /* Blinds_sendEOF */ 
 
/***************************************************************************** 
	Scan_inside() 
*****************************************************************************/ 
int Scan_inside(int end_sym,CONST char* b) 
/*	on entry      : end_sym -> symbolnumber of the expected symbol 
			b 	-> buffer with the HTML-strings 
	on exit       : 0 	-> end_sym found, not end of bufffer b
			1 	-> end sym found, end of buffer b
			2 	-> end sym not found, end of buffer b
			3	-> buffer buf overflow
	description   :	scans the buffer until end of buffer
	variables     : i -> returnvalue from GetString			 
*/ 
{ 
  int i=0; 
 
  if (sym!=end_sym) { 
    if (sym==LNK) { 
      strncpy(help_buf,buf,MAX_BUF); 
    } else { 
      if (sym==PUT || (sym==IMG && end_sym==LNK1)) { 
        strncat(help_buf,buf,MAX_BUF); 
      } 
    } 
    i = GetString(b); 
    sym = GetSym(); 
    while (i==0 && sym!=end_sym) { 
      if (sym==PUT || (sym==IMG && end_sym==LNK1)) {   
        strncat(help_buf,buf,MAX_BUF); 
      } 
      i = GetString(b); 
      sym = GetSym(); 
    } 
    if (sym==end_sym) { 
      strncat(help_buf,buf,MAX_BUF), 
      strncpy(buf,help_buf,MAX_BUF); 
      help_buf[0]='\0'; 
      if (i==0) { 
        return 0; 
      } else { 
        return 1; 
      } 
    } else if (i==1) { 
      if (sym==PUT || (sym==IMG && end_sym==LNK1)) {   
        strncat(help_buf,buf,MAX_BUF); 
      } 
      return 2; 
    } else { 
      return i; 
    } 
  } else { 
    strncat(help_buf,buf,MAX_BUF), 
    strncpy(buf,help_buf,MAX_BUF); 
    help_buf[0]='\0'; 
    return 0; 
  } 
} /* end Scan_inside */

/***************************************************************************** 
	Scan_Link() 
*****************************************************************************/
int Scan_Link(HTStream* me,CONST char* b) 
/*	on entry      : me -> HTBlinds-Stream
			b  -> buffer with the HTML-strings
	on exit       : 0  -> not end of buffer b
                        1  -> end of buffer b			 
	description   :	starts Scan_inside
	variables     : k  -> returnvalue from Scan_inside
*/ 
{
  int k;

  k= Scan_inside(LNK1,b); 
  if (k==0) {
    Reader(LNK,me,buf); 
    awaited_sym=0;
    return 0; 
  } else if (k==1) { 
    Reader(LNK,me,buf); 
    awaited_sym=0; 
    return 1; 
  } else if (k==2) { 
    awaited_sym=LNK1; 
    return 1; 
  } else {
    Reader(PUT,me,help_buf);
    help_buf[0]='\0';
    Reader(PUT,me,buf);
    awaited_sym=0;
    return 0;
  }
} /* end Scan_Link */

/***************************************************************************** 
	Scan_Titel() 
*****************************************************************************/ 
int Scan_Titel(HTStream* me,CONST char* b) 
/*	on entry      : me -> HTBlinds-Stream
			b  -> buffer with the HTML-strings
	on exit       : 0 -> not end of buffer b
                        1 -> end of buffer b			 
	description   :	read the HTML-Buffer until the symbol TIT1 
			is recognized or end of buffer 
	variables     : k -> returnvalue from Scan_inside
*/ 
{
  int i=0,k=0;

  if (sym!=awaited_sym) { 
    if (sym==TIT) { 
      strncpy(help_buf,buf,MAX_BUF); 
      help_tit[0]= '\0'; 
    } else { 
      if (sym==PUT) { 
        strncat(help_buf,buf,MAX_BUF); 
        strncat(help_tit,buf,MAX_BUF); 
       } 
    } 
    i = GetString(b); 
    sym = GetSym(); 
    while (i==0 && sym!=TIT1) { 
      if (sym==PUT) { 
        strncat(help_buf,buf,MAX_BUF); 
        strncat(help_tit,buf,MAX_BUF); 
      } else if (sym==LNK) { 
        Reader(PUT,me,help_tit); 
        help_tit[0]='\0'; 
        return(Scan_Link(me,b));
      } 
      i = GetString(b); 
      sym = GetSym(); 
    } /* end while */ 
  } /* end if */
  if (sym==TIT1) { 
    strncat(help_buf,buf,MAX_BUF); 
    Reader(TIT,me,help_buf);  
    help_buf[0]='\0'; 
    help_tit[0]='\0'; 
    awaited_sym=0; 
    if (i==1) { 
      return 1; 
    } else {
      return 0;
    }
  } else if (i==1) { 
    if (sym==PUT) { 
      strncat(help_buf,buf,MAX_BUF); 
      strncat(help_tit,buf,MAX_BUF); 
    } 
    return 1; 
  } else if (i==2) { 
    return 1; 
  } else {
    Reader(PUT,me,help_buf);
    help_buf[0]='\0';
    Reader(PUT,me,buf);
    awaited_sym=0;
    return 0;
  }
} /* end Scan_Titel */

/***************************************************************************** 
	Blinds_send() 
*****************************************************************************/ 
void Blinds_send(HTStream* me,CONST char* b) 
/*	on entry      : me -> HTBlinds-Stream 
			b  -> buffer with the HTML-strings 
	on exit       : void			 
	description   :	read the HTML-Buffer until a symbol 
			is recognized or end of buffer 
			if end of buffer is arrived then 
			set status according to completeness 
			of read symbol
	variables     : i -> returnvalue from GetString
			k -> returnvalue from Scan_inside
*/ 
{ 
  int i=0,k=0; 
 
  /* for logfile
  if (file_open==0) { 
    fp_scan=fopen("scan.log","w"); 
    file_open=1; 
  } 
  */
 
  InitSym(); 
  while ((i=GetString(b))==0 || i==3) { 
    sym=GetSym(); 
    if ((awaited_sym==0)&&(sym!=TIT)&&(sym!=LNK)&& 
        (sym!=TIT1)&&(sym!=LNK1)) { 
      Reader(sym,me,buf);
    } else { 
      if ((sym==LNK)||(awaited_sym==LNK1)) { 
        if (Scan_Link(me,b)==1) {
          return;
        }
      } else if ((sym==TIT)||(awaited_sym==TIT1)) { 
        awaited_sym=TIT1;
        if (Scan_Titel(me,b)==1) {
          return;
        }
      } 
    } /* if */ 
  } /* end while */ 
  if (awaited_sym==0 && i==1) { 
    sym=GetSym(); 
    Reader(sym,me,buf); 
  } 
} /* end Blinds_send */ 

/***************************************************************************** 
	end procedures & functions 
*****************************************************************************/
