#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include "reader.h"
#include "HTStream.h"
#include "HTBlinds.h"

#define false 0
#define true 1
#define max_strlen 256
#define max_attrlen 256
#define max_txtlen 256
#define max_bufsize 8192
#define max_header 6
#define nr_trail 60

#define link_title "Titellink"
#define link_subs "Subtitellink"
#define link_linklist "Linklist"

#define format_tit "<br><a name=\"%s\">%s</a>%s<br><a href=\"#"link_subs"%i\">%s%s</a><br>"
#define format_lnk "%s%s%s%s"
#define format_beg "\n<a name=\""link_title"0\" href=\"#"link_linklist"\">%s%s</a><br><a href=\"#"link_subs"0\">%s%s</a><br>\n"
#define format_lls_top "<br>\n<a name=\""link_linklist"\">%i%s</a><br>\n"
#define format_lls_item "<a href=\"%s\">%s%s</a><br>\n"
#define format_tls_top "<br>\n<a name=\""link_subs"%i\">%i%s%s</a><br>\n"
#define format_tls_item "<a href=\"#%s\">%s%s</a><br>\n"
#define format_tls_up "<a href=\"#"link_subs"%i\">%s%s</a><br>\n"
#define format_tls_back "<a href=\"#%s\">%s%s</a><br>\n"
#define format_tls_down "<a href=\"#"link_subs"%i\"> %s%s</a><br>\n"
#define format_top "<a href=\"#"link_title"0\">%s%s</a><br>\n"
#define format_img "%s\"%s\""
#define format_bup "<br>\n"
#define format_edc "\n<br>%s<br>\n"
#define format_ell "\n<br>%s<br>\n"
#define format_bdy "<body>\n"
#define format_p "<p>\n"

#define cfgfile_txt "tf.text.config"
#define cfgfile_prod "tf.prod.config"

typedef unsigned char boolean;

typedef struct link_tag {
	char href[max_attrlen];
	char text[max_txtlen];
	struct link_tag *next;
} Link;

typedef struct titel_tag {
	int h; /* nr of the HTML header */
	int y; /* actual level of the titel */
	int nr; /* nr of the titel (titels are counted) */
	int ns; /* nr of subtitels */
	char name[max_attrlen]; /* the HTML name attribute */
	char text[max_txtlen]; /* content of the titel */
	Link *link; /* list of links */
	struct titel_tag *next, *sub, *par; /* pointer to next titel in same level, to subtitels and to parent titel */	
} Titel;

char
	txt_link[max_strlen] = "Link ",
	txt_addlink[max_strlen] = "Ref to ",
	txt_h[max_header][max_strlen] =
		{ "Titel level 1 ",
		  "Titel level 2 ", 
		  "Titel level 3 ",
		  "Titel level 4 ",
		  "Titel level 5 ",
		  "Titel level 6 " },
	txt_sub[max_header][max_strlen] =
		{ "Titel level 1",
		  "Titel level 2",
		  "Titel level 3",
		  "Titel level 4",
		  "Titel level 5",
		  "Titel level 6" },
	txt_hlist[max_header][max_strlen] = 
		{ " Titel level 1 ",
		  " Titel level 2 ",
		  " Titel level 3 ",
		  " Titel level 4 ",
		  " Titel level 5 ",
		  " Titel level 6 " },
	txt_tolinks[max_strlen] = "Linklist",
	txt_linklist[max_strlen] = " Links",
	txt_back[max_strlen] = "Back to text",
	txt_up[max_strlen] = "Upper level",
	txt_down[max_strlen] = "Lower level",
	txt_image[max_strlen] = "Image ",
	txt_radio[max_strlen] = "Radiobutton ",
	txt_check[max_strlen] = "Checkbox ",
	txt_edit[max_strlen] = "Editbox ",
	txt_combo[max_strlen] = "Combobox ",
	txt_button[max_strlen] = "Button ",
	txt_endlinks[max_strlen] = "End of linklist",
	txt_enddoc[max_strlen] = "End of document",
	txt_top[max_strlen] = "Top of document";

HTStream *hs;

/* 
 * The following procedures are needed to parse incoming tags and elements.
 */

#define name 0
#define value 1
#define equal 2
#define opn 3
#define cls 4
#define end 5

char text[max_strlen]; /* contains name or textual attribute */

/*
 * This function scans string s beginning at position i.
 * It yields a symbolnumber and sets the text variable
 */

int get_sym(char *s, int *i)
{
	int l, j;
	char q;

	l = strlen(s);
	while (*i<l && isspace(s[*i])) (*i)++;
	if (*i == l) return end;
 	if (s[*i] == '>') {
		(*i)++;
		return cls;
	}
	if (s[*i] == '<' && isalpha(s[(*i)+1])) {
		(*i)++;
		return opn;
	}
	if (s[*i] == '=') {
		(*i)++;
		return equal;
	}
	if (s[*i] == '"' || s[*i] == '\'') {
		q = s[*i];
		(*i)++;
		j = 0;
		while (*i<l && s[*i] != q) {
			text[j] = s[*i];
			(*i)++; j++;
		}
		text[j] = 0;
		if (*i<l) (*i)++;
		return value;
	}
	j = 0;
	do {
		text[j] = s[*i];
		(*i)++; j++;
	} while (*i<l && !isspace(s[*i]) && s[*i] != '>' && s[*i] != '=');
	text[j] = '\0';
	return name;
}

/*
 * This function gets the value of attribute aname in tag and returns it in rval.
 * If the attribute is found the function returns 0, else 1.
 */

int get_attribute(char *tag, char *aname, char *rval)
{
	int i, sym;
	
	i = 0; rval[0] = '\0';
	sym = get_sym(tag, &i);
	while (sym != end && sym != cls && (sym != name || strcasecmp(aname, text) != 0)) sym = get_sym(tag, &i);
	if (sym == end || sym == cls) return 1;
	sym = get_sym(tag, &i);
	if (sym != equal) return 0;
	sym = get_sym(tag, &i);
	strcpy(rval, text);
	return 0;
}

/*
 * This function subdivides an element into starttag, middle part and endtag.
 */

void subdivide_element(char *elem, char *st, char *mid, char *et)
{
	int a, b, c, i, j, sym;

	c = strlen(elem);
	b = c;
	if (elem[b-1] == '>') while (b>0 && elem[b] != '<') b--;
	a = 1;
	sym = get_sym(elem, &a);
	while (sym != cls && sym != end) sym = get_sym(elem, &a);
	
	for (i=0; i<a; i++) st[i]=elem[i];
	st[i]='\0';
	for (i=a, j=0; i<b; i++, j++) mid[j]=elem[i];
	mid[j]='\0';
	for (i=b, j=0; i<c; i++, j++) et[j]=elem[i];
	et[j]='\0'; 				
}

/*
 * This function locates an IMG element in in_text and returns startposition a and endposition b.
 * The function returns 0 if the element is found, 1 else.
 */

int locate_imgelem(char *in_text, int *a, int *b)
{
	int i, sym;

	i = 0;
	sym = get_sym(in_text, &i);
	while (true) {
		while (sym != end && sym != opn) sym = get_sym(in_text, &i);
		if (sym == end) return 1;
		*a = i - 1;
		sym = get_sym(in_text, &i);
		if (sym == name && strcasecmp(text, "img") == 0) {
			while (sym != end && sym != cls) sym = get_sym(in_text, &i);
			*b = i;
			return 0;
		}
		sym = get_sym(in_text, &i);
	}	
}

/*
 * This function returns 1 if a string is only containing space, 0 else.
 */

int is_empty(char *text)
{
	int i;

	for (i=0; i<strlen(text); i++) if (!isspace(text[i])) return 0;
	return 1;
}

/* 
 * This function extracts the headernr out of the HTML headertag st and returns it.
 */

int get_headernr(char *st)
{
	int i, nr;
	char ns[2];

	i = 0;
	get_sym(st, &i);
	if (get_sym(st, &i) == name && isdigit(text[1])) {
		ns[0] = text[1]; ns[1] = '\0';
		nr = atoi(ns);
		return nr;
	}
	return max_header;
}

#undef name
#undef value
#undef equal
#undef opn
#undef cls
#undef end

/* 
 * This function sends a string to HTBlinds, which sends it to the client browser.
 */

int outf(char *fmt, ...)
{
	va_list ap;
	char *p, *sval;
	int ival, n;
	char buf[max_bufsize], s[max_strlen];

	n = 0;
	va_start(ap, fmt);
	for (p = fmt; *p; p++) {
		if (*p == '%') switch (*++p) {
			case 'd':
			case 'i':
				ival = va_arg(ap, int);
				sprintf(s, "%i", ival);
				buf[n] = '\0';
				strcat(buf, s);
				n = strlen(buf);
				break;
			case 's':
				sval = va_arg(ap, char *);
				buf[n] = '\0';
				strcat(buf, sval);
				n = strlen(buf);
				break;
			default:
				buf[n++] = *p;
		}
		else if (*p == '\\') switch (*++p) {
			case 'n': buf[n++] = '\n'; break;
			case 't': buf[n++] = '\t'; break;
			default: buf[n++] = *p;
		}
		else buf[n++] = *p;
	}
	va_end(ap);
	buf[n] = '\0';
	Blinds_receive(hs, buf, strlen(buf));
	return 0;
}

/*
 * The following functions are needed to control titels and links.
 */

Titel *first = NULL; /* points to the first titel of the titel structure */
Titel *curt; /* points to the actual titel */
Link *curl; /* points to the actual link */
int tnr = 0; /* titels are counted */
int lnr = 0; /* links are counted */

/*
 * This function inserts a titel of HTML level h, NAME attribute aname and content text into the titel structure.
 * It returns 0 if ok, 1 else.
 */

int insert_titel(int h, char *name, char *text)
{
	Titel *t;
	
	t = (Titel *)malloc(sizeof(Titel));
	if (t == NULL) return 1;
	t->h = h;
	if (strlen(name) == 0) sprintf(t->name, link_title"%i", tnr);
	else {
		strncpy(t->name, name, max_attrlen); t->name[max_attrlen-1] = '\0';
	}
	strncpy(t->text, text, max_txtlen); t->text[max_txtlen-1] = '\0';
	t->sub = NULL;
	t->next = NULL;
	t->link = NULL;
	t->nr = tnr++;
	t->ns = 0;
	curl = NULL;

	if (first) {
		if (h > curt->h) {
			curt->sub = t;
			t->par = curt;
		}
		else if (h == curt->h) {
			curt->next = t;
			t->par = curt->par;
		}
		else {
			while (curt->par && curt->par->h >= h) curt = curt->par;
			t->par = curt->par;
			curt->next = t;
		}
		curt = t;
		t->y = t->par->y + 1;
	}
	else {
		t->par = NULL; t->y = 0;
		first = t; curt = t;
	}
	if (curt->par) curt->par->ns++;
	return 0;
}

/*
 * This function inserts a link with HREF attribute href and content text into the link structure.
 * It returns 0 if ok, 1 else.
 */

int insert_link(char *href, char *text)
{
	Link *l;

	l = (Link *)malloc(sizeof(Link));
	if (l == NULL) return 1;

	strncpy(l->href, href, max_attrlen); l->href[max_attrlen] = '\0';
	strncpy(l->text, text, max_txtlen); l->text[max_txtlen] = '\0';
	l->next = NULL;
	if (curl) curl->next = l;
	else curt->link = l;
	curl = l;
	lnr++;
	return 0;
}

/*
 * This function returns allocated memory of Titel t with all subtitels, consecutive titels and links.
 */

void delete_titel(Titel *t)
{
	Link *l0, *l1;
	
	if (t->sub == NULL && t->next == NULL) {
		l0 = t->link;
		while (l0) {
			l1 = l0;
			l0 = l0->next;
			free(l1);
		}		
		free(t);
	}
	else {
		if (t->sub) delete_titel(t->sub);
		if (t->next) delete_titel(t->next);
	}
}

/*
 * This function prints out the link list.
 */

int list_links(Titel *t)
{	
	Link *l;	

	if (t) {
		l = t->link;
		while (l) {
			outf(format_lls_item, l->href, txt_link, l->text);
			l = l->next;
		}
		list_links(t->sub);
		list_links(t->next);
	}
	return 0;
}

/*
 * This function prints out the titel list.
 */

int list_titels(Titel *t)
{
	Titel *s;

	if (t) {
		outf(format_tls_top, t->nr, t->ns, txt_hlist[t->y], t->text);
		s = t->sub;
		while (s) {
			outf(format_tls_item, s->name, txt_addlink, s->text);
			outf(format_tls_down, s->nr, txt_addlink, txt_down);
			s = s->next;
		}
		if (t->par) outf(format_tls_up, t->par->nr, txt_addlink, txt_up);
		outf(format_tls_back, t->name, txt_addlink, txt_back);
		outf(format_top, txt_addlink, txt_top);
		list_titels(t->sub);
		list_titels(t->next);
	}
	return 0;
}

/*
 * The following functions deal with the productions.
 */


void prod_img0(char *in_text, char *out_text)
{
	char srcs[max_strlen];

	if (get_attribute(in_text, "src", srcs) == 1) strcpy(srcs, "");
	sprintf(out_text, format_img, txt_image, srcs);
}

int prod_img(char *elem)
{
	char out[max_strlen];
	
	prod_img0(elem, out);
	return outf("%s", out);	
}

int prod_lnk(char *elem)
{
	char href[max_strlen], text[max_strlen], text1[max_strlen], st[max_strlen], mid[max_strlen], et[max_strlen];
	int a, b, i, j;

	subdivide_element(elem, st, mid, et);		
	if (get_attribute(st, "href", href) == 1) return 1;
	if (locate_imgelem(mid, &a, &b) == 0) {
		for (i=a, j=0; i<b; i++, j++) text[j] = mid[i];
		text[j] = '\0';
		prod_img0(text, text1);
		for (i=0, j=0; i<a; i++, j++) text[j] = mid[i];
		for (i=0; i<strlen(text1); i++, j++) text[j] = text1[i];
		for (i=b; i<strlen(mid); i++, j++) text[j] = mid[i];
		text[j] = '\0';
	}
	else strcpy(text, mid);
	if (insert_link(href, text) == 1) return 1;
	outf(format_lnk, st, txt_link, text, et);
	return 0;
}

int prod_tit(char *elem)
{
	char st[max_strlen], mid[max_strlen], et[max_strlen], id[max_strlen];
	int h;

	subdivide_element(elem, st, mid, et);
	if (is_empty(mid)) return 0;
	h = get_headernr(st);
	get_attribute(st, "id", id);
	if (insert_titel(h, id, mid) == 1) return 1;
	outf(format_tit, curt->name, txt_h[curt->y-1], mid, curt->nr, txt_addlink, txt_sub[curt->y]);
	return 0;
}

int prod_beg(void)
{
	outf(format_beg, txt_addlink, txt_tolinks, txt_addlink, txt_sub[0]);
	return 0;
}

int prod_tls(void)
{
	outf("\n");
	return list_titels(first);
}

int prod_lls(void)
{
	outf(format_lls_top, lnr, txt_linklist);
	list_links(first);
	outf(format_top, txt_addlink, txt_top);
	return 0;
}

int prod_edc(void)
{
	outf(format_edc, txt_enddoc);
	return 0;
}

int prod_ell(void)
{
	outf(format_ell, txt_endlinks);
	return 0;
}

int prod_blu(void)
{
	int i;

	for (i=0; i<nr_trail; i++) outf(format_bup);
	return 0;
}

int prod_bdy(char *text)
{
	outf(format_bdy);
	return 0;
}

int prod_p(char *text)
{
	outf(format_p);
	return 0;
}

/*
 * The following functions are used to configure productions and texts.
 */

boolean
	adr_on = true,
	blk_on = true,
	tt_on = true,
	b_on = true,
	i_on = true,
	u_on = true,
	em_on = true,
	str_on = true,
	dfn_on = true,
	cit_on = true,
	var_on = true,
	rad_on = true,
	chk_on = true,
	edt_on = true,
	cbb_on = true,
	but_on = true,
	img_on = true,
	lnk_on = true,
	tit_on = true,
	beg_on = true,
	tls_on = true,
	lls_on = true,
	edc_on = true,
	ell_on = true,
	blu_on = true,
	bdy_on = true,
	ft_on = true,
	cen_on = true,
	s_on = true,
	big_on = true,
	sml_on = true,
	p_on = true;

/*
 * This function reads a line out of configuration file fp.
 */

int get_cfline(FILE *fp, char *line)
{
	char *res;

	res = fgets(line, max_strlen, fp);
	while (res != NULL && line[0] == ';') res = fgets(line, max_strlen, fp);
	if (res) {
		line[strlen(line)-1] = '\0';
		return 0;
	}
	else return 1;
}

/*
 * This function configures the texts.
 */

int config_txt(void)
{
	FILE *fp;
	int i;

	if ((fp = fopen(cfgfile_txt, "r")) != NULL) {
		get_cfline(fp, txt_link);
		get_cfline(fp, txt_addlink);
		for (i=0; i<max_header; i++) get_cfline(fp, txt_h[i]);
		for (i=0; i<max_header; i++) get_cfline(fp, txt_sub[i]);
		for (i=0; i<max_header; i++) get_cfline(fp, txt_hlist[i]); 
		get_cfline(fp, txt_tolinks);
		get_cfline(fp, txt_linklist);
		get_cfline(fp, txt_back);
		get_cfline(fp, txt_up);
		get_cfline(fp, txt_down);
		get_cfline(fp, txt_image);
		get_cfline(fp, txt_radio);
		get_cfline(fp, txt_check);
		get_cfline(fp, txt_edit);
		get_cfline(fp, txt_combo);
		get_cfline(fp, txt_button);
		get_cfline(fp, txt_endlinks);
		get_cfline(fp, txt_enddoc);
		return get_cfline(fp, txt_top);
	}
	return 1;
}

/*
 * This function switches a production on or off
 */

int set_prod(FILE *fp, boolean *prod)
{
	char line[max_strlen];

	if (get_cfline(fp, line) == 0) {
		if (strchr(line, '+') != NULL) *prod = true;
		else *prod = false;
		return 0;
	}
	return 1;
}

/*
 * This function configures the productions.
 */

int config_prod(void)
{
	FILE *fp;

	if ((fp = fopen(cfgfile_prod, "r")) != NULL) {
		set_prod(fp, &adr_on);
		set_prod(fp, &blk_on);
		set_prod(fp, &tt_on);
		set_prod(fp, &b_on);
		set_prod(fp, &i_on);
		set_prod(fp, &u_on);
		set_prod(fp, &em_on);
		set_prod(fp, &str_on);
		set_prod(fp, &dfn_on);
		set_prod(fp, &cit_on);
		set_prod(fp, &var_on);
		set_prod(fp, &cen_on);
		set_prod(fp, &s_on);
		set_prod(fp, &big_on);
		set_prod(fp, &sml_on);
		set_prod(fp, &bdy_on);
		set_prod(fp, &p_on);
		set_prod(fp, &ft_on);
		set_prod(fp, &rad_on);
		set_prod(fp, &chk_on);
		set_prod(fp, &edt_on);
		set_prod(fp, &cbb_on);
		set_prod(fp, &but_on);
		set_prod(fp, &img_on);
		set_prod(fp, &lnk_on);
		set_prod(fp, &tit_on);
		set_prod(fp, &beg_on);
		set_prod(fp, &tls_on);
		set_prod(fp, &lls_on);
		set_prod(fp, &edc_on);
		set_prod(fp, &ell_on);
		return set_prod(fp, &blu_on);
	}
	return 1;
}

/*
 * The following functions are called from other parts.
 */

int init_transform(void)
{
	if (first) delete_titel(first);
	tnr = 0; lnr = 0;
	first = NULL; curl = NULL; curt = NULL;
	config_txt();
	config_prod();
	return insert_titel(0, "", "");
}

int transform(int prod, char *text, HTStream *me)
{
	hs = me;
	switch (prod) {
		case ADR0:
		case ADR1: if (adr_on) return 0; break;
		case BLK0:
		case BLK1: if (blk_on) return 0; break;
		case TT0:
		case TT1: if (tt_on) return 0; break;
		case B0:
		case B1: if (b_on) return 0; break;
		case I0:
		case I1: if (i_on) return 0; break;
		case U0:
		case U1: if (u_on) return 0; break;
		case EM0:
		case EM1: if (em_on) return 0; break;
		case STR0:
		case STR1: if (str_on) return 0; break;
		case DFN0:
		case DFN1: if (dfn_on) return 0; break;
		case CIT0:
		case CIT1: if (cit_on) return 0; break;
		case VAR0:
		case VAR1: if (var_on) return 0; break;
		case FT0:
		case FT1: if (ft_on) return 0; break;
		case CEN0:
		case CEN1: if (cen_on) return 0; break;
		case S0:
		case S1: if (s_on) return 0; break;
		case BIG0:
		case BIG1: if (big_on) return 0; break;
		case SML0:
		case SML1: if (sml_on) return 0; break;
		case RAD: if (rad_on) return outf("%s", txt_radio); break;
		case CHK: if (chk_on) return outf("%s", txt_check); break;
		case EDT: if (edt_on) return outf("%s", txt_edit); break;
		case CBB: if (cbb_on) return outf("%s", txt_combo); break;
		case BUT: if (but_on) return outf("%s", txt_button); break;
		case IMG: if (img_on) return prod_img(text); break;
		case LNK: if (lnk_on) return prod_lnk(text); break;
		case TIT: if (tit_on) return prod_tit(text); break;
		case BDY: if (bdy_on) return prod_bdy(text); break;
		case P: if (p_on) return prod_p(text); break;
		case BEG: if (beg_on) return prod_beg(); break;
		case TLS: if (tls_on) return prod_tls(); break;
		case LLS: if (lls_on) return prod_lls(); break;
		case EDC: if (edc_on) return prod_edc(); break;
		case ELL: if (ell_on) return prod_ell(); break;
		case BLU: if (blu_on) return prod_blu(); break;
	}
	return outf("%s", text);
}

void close_transform(void)
{
	if (first) delete_titel(first);
}
